/*
 * RFC2367 PF_KEYv2 Key management API domain socket I/F
 * Copyright (C) 1999  Richard Guy Briggs.
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * RCSID $Id: pfkey_v2.c,v 1.11 1999/09/30 02:55:10 rgb Exp $
 */

/*
 *		Template from /usr/src/linux-2.0.36/net/unix/af_unix.c.
 *		Hints from /usr/src/linux-2.0.36/net/ipv4/udp.c.
 */

/*
 * NET3:	Implementation of BSD Unix domain sockets.
 *
 * Authors:	Alan Cox, <alan@cymru.net>
 *
 *		Currently this contains all but the file descriptor passing code.
 *		Before that goes in the odd bugs in the iovec handlers need 
 *		fixing, and this bit testing. BSD fd passing is not a trivial part
 *		of the exercise it turns out. Anyone like writing garbage collectors.
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *
 * Fixes:
 *		Linus Torvalds	:	Assorted bug cures.
 *		Niibe Yutaka	:	async I/O support.
 *		Carsten Paeth	:	PF_UNIX check, address fixes.
 *		Alan Cox	:	Limit size of allocated blocks.
 *		Alan Cox	:	Fixed the stupid socketpair bug.
 *		Alan Cox	:	BSD compatibility fine tuning.
 *		Alan Cox	:	Fixed a bug in connect when interrupted.
 *		Alan Cox	:	Sorted out a proper draft version of
 *					file descriptor passing hacked up from
 *					Mike Shaver's work.
 *		Marty Leisner	:	Fixes to fd passing
 *		Nick Nevin	:	recvmsg bugfix.
 *		Alan Cox	:	Started proper garbage collector
 *		Heiko EiBfeldt	:	Missing verify_area check
 *		Alan Cox	:	Shutdown() bug
 *
 * Known differences from reference BSD that was tested:
 *
 *	[TO FIX]
 *	ECONNREFUSED is not returned from one end of a connected() socket to the
 *		other the moment one end closes.
 *	fstat() doesn't return st_dev=NODEV, and give the blksize as high water mark
 *		and a fake inode identifier (nor the BSD first socket fstat twice bug).
 *	[NOT TO FIX]
 *	accept() returns a path name even if the connecting socket has closed
 *		in the meantime (BSD loses the path and gives up).
 *	accept() returns 0 length path for an unbound connector. BSD returns 16
 *		and a null first byte in the path (but not for gethost/peername - BSD bug ??)
 *	socketpair(...SOCK_RAW..) doesn't panic the kernel.
 *	BSD af_unix apparently has connect forgetting to block properly.
 */

#define __NO_VERSION__
#include <linux/module.h>
#include <linux/version.h>

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/stat.h>
#include <linux/socket.h>
#include <linux/un.h>
#include <linux/fcntl.h>
#include <linux/termios.h>
#include <linux/socket.h>
#include <linux/sockios.h>
#include <linux/net.h>
#include <linux/in.h>
#include <linux/fs.h>
#include <linux/malloc.h>
#include <asm/segment.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <net/sock.h>
/* #include <net/tcp.h> */
#include <net/af_unix.h>
#include <linux/proc_fs.h>

#include <linux/types.h>
 
#include "../../../lib/freeswan.h"
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
#include <asm/uaccess.h>
#include <linux/in6.h>
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
#include "pfkey.h"
#include "pfkeyv2.h"
#include "radij.h"
#include "ipsec_encap.h"
#include "ipsec_netlink.h"

#ifdef DEBUG_IPSEC
int debug_pfkey = 0;
#endif

#define min(a,b)	(((a)<(b))?(a):(b))

#define SENDERR(_x) do { len = -(_x); goto errlab; } while (0)

#if 0 && (LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0))
static atomic_t pfkey_nr_socks = ATOMIC_INI(0);
#endif /* 0 && LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */

pfkey_sock *pfkey_sock_list=NULL;
int pfkey_msg_parse(struct sock *, struct sadb_msg *);

#if 0 && LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
static void pfkey_destruct(struct sock *sk)
{
	return;
}
#endif /* 0 && LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
static void
pfkey_remove_socket(pfkey_sock *sk)
{
	pfkey_sock **s;
	
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_remove_socket: .\n");
	cli();
	s=&pfkey_sock_list;

	while(*s!=NULL) {
		if(*s==sk) {
			*s=sk->next;
			/* XXX rgb */ sk->next=NULL;
			sti();
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_remove_socket: succeeded.\n");
			return;
		}
		s=&((*s)->next);
	}
	sti();
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_remove_socket: not found.\n");
	return;
}

static void
pfkey_insert_socket(pfkey_sock *sk)
{
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_insert_socket: .\n");
	cli();
	sk->next=pfkey_sock_list;
	pfkey_sock_list=sk;
	sti();
}
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0) */

static void
pfkey_destroy_socket(pfkey_sock *sk)
{
#if 1 || LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
	struct sk_buff *skb;
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0) */

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_destroy_socket: .\n");
#if 0 && LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
	sklist_destroy_socket(&pfkey_sock_list, sk);
#else /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
	sklist_remove_socket(&pfkey_sock_list, sk);
#else /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
	pfkey_remove_socket(sk);
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
	
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_destroy_socket: sk=%p, sk->receive_queue=%p, &sk->receive_queue=%p.\n",
		    sk,
		    &(sk->receive_queue),
		    &(sk->receive_queue));

	while((sk && &(sk->receive_queue)) && ((skb=skb_dequeue(&(sk->receive_queue)))!=NULL)) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_destroy_socket: skb=%p freed.\n", skb);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
		if((void*)skb == (void*)0x8) {
			KLIPS_PRINT(debug_pfkey,
				    "klips_debug:pfkey_destroy_socket: Why is this skb=%p?\n", skb);
		} else {
			kfree_skb(skb);
		}
#else /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
		kfree_skb(skb, FREE_WRITE);
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
	}
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_destroy_socket: all skbs dequeued and freed.\n");

	sk_free(sk); 
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_destroy_socket: destroyed.\n");
}

static int
pfkey_fcntl(struct socket *sock, unsigned int cmd, unsigned long arg)
{
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_fcntl: not supported.\n");
	return -EINVAL;
}

static int
pfkey_setsockopt(struct socket *sock, int level, int optname, char *optval, int optlen)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
	pfkey_sock *sk;

	if(sock == NULL) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_setsockopt: Null socket passed in.\n");
		return -EINVAL;
	}
	
	sk=sock->data;
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0) */

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_setsockopt: .\n");
	if(level!=SOL_SOCKET) {
		return -EOPNOTSUPP;
	}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
	return sock_setsockopt(sock, level, optname, optval, optlen);
#else /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
	return sock_setsockopt(sk, level, optname, optval, optlen);
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
}

static int
pfkey_getsockopt(struct socket *sock, int level, int optname, char *optval, int *optlen)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
	pfkey_sock *sk;

	if(sock == NULL) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_setsockopt: Null socket passed in.\n");
		return -EINVAL;
	}
	
	sk=sock->data;
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0) */

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_getsockopt: .\n");
	if(level!=SOL_SOCKET) {
		return -EOPNOTSUPP;
	}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
	return sock_getsockopt(sock, level, optname, optval, optlen);
#else /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
	return sock_getsockopt(sk, level, optname, optval, optlen);
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
}

static int
pfkey_listen(struct socket *sock, int backlog)
{
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_listen: not supported.\n");
	return -EINVAL;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
static void
pfkey_state_change(struct sock *sk)
{
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_state_change: .\n");
	if(!sk->dead) {
		wake_up_interruptible(sk->sleep);
	}
}

static void
pfkey_data_ready(struct sock *sk, int len)
{
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_data_ready: .\n");
	if(!sk->dead) {
		wake_up_interruptible(sk->sleep);
		sock_wake_async(sk->socket, 1);
	}
}

static void
pfkey_write_space(struct sock *sk)
{
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_write_space: .\n");
	if(!sk->dead) {
		wake_up_interruptible(sk->sleep);
		sock_wake_async(sk->socket, 2);
	}
}
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0) */

struct proto_ops pfkey_dgram_ops;

static int
pfkey_create(struct socket *sock, int protocol)
{
	struct sock *sk;

	if(sock == NULL) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_create: socket NULL.\n");
		return -EINVAL;
	}

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_create: type:%d state:%d flags:%ld protocol:%d\n",
		    sock->type, (unsigned int)(sock->state), sock->flags, protocol);

	if(sock->type != SOCK_RAW) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_create: only SOCK_RAW supported.\n");
		return -ESOCKTNOSUPPORT;
	}

	if(protocol != PF_KEY_V2) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_create: protocol not PF_KEY_V2.\n");
		return -EPROTONOSUPPORT;
	}

	if((current->uid != 0)) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_create: must be root to open pfkey sockets.\n");
		return -EACCES;
	}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
	sock->state = SS_UNCONNECTED;
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
	MOD_INC_USE_COUNT;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
	if((sk=(pfkey_sock *)sk_alloc(PF_KEY, GFP_KERNEL, 1)) == NULL)
#else /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
	if((sk=(pfkey_sock *)sk_alloc(GFP_KERNEL)) == NULL)
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
	{
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_create: Out of memory trying to allocate.\n");
		MOD_DEC_USE_COUNT;
		return -ENOMEM;
	}

	memset(sk, 0, sizeof(*sk));

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
	sk->reuse = 1;
	sock->ops = &pfkey_dgram_ops;
	sock_init_data(sock, sk);

	sk->zapped=0;
	sk->family = PF_KEY;
	sk->num = protocol;
	sklist_insert_socket(&pfkey_sock_list, sk);
#else /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
	sk->type=sock->type;
	init_timer(&sk->timer);
	skb_queue_head_init(&sk->write_queue);
	skb_queue_head_init(&sk->receive_queue);
	skb_queue_head_init(&sk->back_log);
	sk->rcvbuf=SK_RMEM_MAX;
	sk->sndbuf=SK_WMEM_MAX;
	sk->allocation=GFP_KERNEL;
	sk->state=TCP_CLOSE;
	sk->priority=SOPRI_NORMAL;
	sk->state_change=pfkey_state_change;
	sk->data_ready=pfkey_data_ready;
	sk->write_space=pfkey_write_space;
	sk->error_report=pfkey_state_change;
	sk->mtu=4096;
	sk->socket=sock;
	sock->data=(void *)sk;
	sk->sleep=sock->wait;
	pfkey_insert_socket(sk);
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_create: Socket initialised.\n");
	return 0;
}

static int
pfkey_dup(struct socket *newsock, struct socket *oldsock)
{
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_dup: operation not supported.\n");
	return -EINVAL;
}

static int
pfkey_release(struct socket *sock, struct socket *peer)
{
	pfkey_sock *sk;

	if(sock==NULL) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_release: No socket attached.\n");
		return 0;
	}
		
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
	sk=sock->sk;
#else /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
	sk=sock->data;
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
	
	/* May not have data attached */
	if(sk==NULL) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_release: No sock attached.\n");
		return 0;
	}
		
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_release: .\n");

	/* sklist_remove_socket(&pfkey_sock_list, sk); */

	if(!sk->dead)
		sk->state_change(sk);
	sk->dead=1;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
	sock->sk = NULL;
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */

	/* Try to flush out this socket. Throw out buffers at least */
	pfkey_destroy_socket(sk);
	MOD_DEC_USE_COUNT;
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_release: succeeded.\n");
	return 0;
}


static int
pfkey_bind(struct socket *sock, struct sockaddr *uaddr, int addr_len)
{
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_bind: operation not supported.\n");
	return -EINVAL;
}

static int
pfkey_connect(struct socket *sock, struct sockaddr *uaddr, int addr_len, int flags)
{
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_connect: operation not supported.\n");
	return -EINVAL;
}

static int
pfkey_socketpair(struct socket *a, struct socket *b)
{
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_socketpair: operation not supported.\n");
	return -EINVAL;
}

static int
pfkey_accept(struct socket *sock, struct socket *newsock, int flags)
{
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_aaccept: operation not supported.\n");
	return -EINVAL;
}

static int
pfkey_getname(struct socket *sock, struct sockaddr *uaddr, int *uaddr_len,
		int peer)
{
	struct sockaddr *ska = (struct sockaddr*)uaddr;
	
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_getname: .\n");
	ska->sa_family = PF_KEY;
	*uaddr_len = sizeof(*ska);
	return 0;
}

static int
pfkey_upmsg(struct socket *sk, struct sadb_msg *pfkey_msg)
{
	int error;
	struct sk_buff * skb = NULL;
	struct sock *sock;

	if(sk == NULL) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_upmsg: NULL socket passed in.\n");
		return 0;
	}

	if(pfkey_msg == NULL) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_upmsg: NULL pfkey_msg passed in.\n");
		return 0;
	}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
	sock = sk->sk;
#else /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
	sock = sk->data;
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_upmsg: operation not supported.\n");
	if(!(skb = alloc_skb(pfkey_msg->sadb_msg_len, GFP_ATOMIC) )) {
		return -ENOBUFS;
	}

	skb->h.raw = skb_put(skb, pfkey_msg->sadb_msg_len);
	memcpy(skb->h.raw, pfkey_msg, pfkey_msg->sadb_msg_len * IPSEC_PFKEYv2_ALIGN);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
	skb->free = 1;
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0) */

	if((error = sock_queue_rcv_skb(sock, skb)) < 0) {
		skb->sk=NULL;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
		kfree_skb(skb);
#else /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
		kfree_skb(skb, FREE_WRITE);
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
		return -error;
	}
	return 0;
}

/*
 *	Send PF_KEY data down.
 */
		
static int
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
pfkey_sendmsg(struct socket *sock, struct msghdr *msg, int len, struct scm_cookie *scm)
#else /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
pfkey_sendmsg(struct socket *sock, struct msghdr *msg, int len, int nonblock, int flags)
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
{
	struct sock *sk;
	int error = 0;
	struct sockaddr*ska;
#if 0
	struct sk_buff *skb;
#endif /* 0 */
	struct sadb_msg * pfkey_msg = NULL;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
	sk = sock->sk;
#else /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
	sk = sock->data;
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */

	if(sock == NULL) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_sendmsg: Null socket passed in.\n");
		SENDERR(EINVAL);
	}
	
	if(msg == NULL) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_sendmsg: Null msghdr passed in.\n");
		return -EINVAL;
	}

	ska = (struct sockaddr*)msg->msg_name;

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_sendmsg: .\n");
	if(sk->err) {
		return sock_error(sk);
	}

	if((current->uid != 0)) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_sendmsg: must be root to send messages to pfkey sockets.\n");
		SENDERR(EACCES);
	}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
	if(msg->msg_control)
#else /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
	if(flags || msg->msg_control)
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
	{
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_sendmsg: can't set flags or set msg_control.\n");
		SENDERR(EINVAL);
	}
		
	if(sk->shutdown & SEND_SHUTDOWN) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_sendmsg: shutdown.\n");
		send_sig(SIGPIPE, current, 0);
		SENDERR(EPIPE);
	}
	
	if(len < sizeof(struct sadb_msg)) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_sendmsg: bogus msg len of %d, too small.\n", len);
		SENDERR(EINVAL);
	}

	if((pfkey_msg = (struct sadb_msg*)kmalloc(len, GFP_KERNEL)) == NULL) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_sendmsg: memory allocation error.\n");
		SENDERR(ENOBUFS);
	}

	memcpy_fromiovec((void *)pfkey_msg, msg->msg_iov, len);

	if(pfkey_msg->sadb_msg_version != PF_KEY_V2) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_sendmsg: not PF_KEY_V2 msg.\n");
		SENDERR(EINVAL);
	}

	if(len != pfkey_msg->sadb_msg_len * IPSEC_PFKEYv2_ALIGN) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_sendmsg: bogus msg len of %d, not %d byte aligned.\n",
			    len, IPSEC_PFKEYv2_ALIGN);
		SENDERR(EMSGSIZE);
	}

	/* XXX when this becomes a lib, keying daemons must be able to receive errors */
	if(pfkey_msg->sadb_msg_errno) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_sendmsg: errno set to %d.\n",
			    pfkey_msg->sadb_msg_errno);
		SENDERR(EINVAL);
	}

	/* check PID */
	if(pfkey_msg->sadb_msg_pid != current->pid) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_sendmsg: pid (%d) does not equal sending process pid (%d).\n",
			    pfkey_msg->sadb_msg_pid, current->pid);
		SENDERR(EINVAL);
	}

	if(pfkey_msg->sadb_msg_reserved) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_sendmsg: reserved field must be zero, set to %d.\n",
			    pfkey_msg->sadb_msg_reserved);
		SENDERR(EINVAL);
	}
	
	if((pfkey_msg->sadb_msg_type > SADB_MAX) || (!pfkey_msg->sadb_msg_type)){
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_sendmsg: msg type too large or small:%d.\n",
			    pfkey_msg->sadb_msg_type);
		SENDERR(EINVAL);
	}
	
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_sendmsg: msg sent for parsing.\n");
	if((error = pfkey_msg_parse(sk, pfkey_msg))) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_sendmsg: pfkey_msg_parse returns %d.\n",
			    error);
	}
 errlab:
	if(pfkey_msg) {
		pfkey_msg->sadb_msg_errno = error;
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_sendmsg: send up msg.\n");
		pfkey_upmsg(sock, pfkey_msg);
		kfree_s((void*)pfkey_msg, len);
	}
	if(error) {
		return -error;
	}
	return len;
}

/*
 *	Receive PF_KEY data up.
 */
		
static int
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
pfkey_recvmsg(struct socket *sock, struct msghdr *msg, int size, int flags, struct scm_cookie *scm)
#else /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
pfkey_recvmsg(struct socket *sock, struct msghdr *msg, int size, int noblock, int flags, int *addr_len)
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
{
	struct sock *sk;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
	int noblock = flags & MSG_DONTWAIT;
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
	struct sockaddr *ska;
	struct sk_buff *skb;
	int error;
#if 0
	unsigned char *sp;
	int len;
	int num;
	struct iovec *iov=msg->msg_iov;
#endif /* 0 */

	if(sock == NULL) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_recvmsg: Null socket passed in.\n");
		return -EINVAL;
	}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
	sk = sock->sk;
#else /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
	sk = sock->data;
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */

	if(msg == NULL) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_recvmsg: Null msghdr passed in.\n");
		return -EINVAL;
	}

	ska = (struct sockaddr*)msg->msg_name;

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_recvmsg: .\n");
	if(flags) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_sendmsg: flags (%d) not supported.\n",
			    flags);
		return -EOPNOTSUPP;
	}
		
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
	msg->msg_namelen = sizeof(*ska);
#else /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
	if(addr_len) {
		*addr_len = sizeof(*ska);
	}
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
		
	if(sk->err) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_sendmsg: sk->err=%d.\n", sk->err);
		return sock_error(sk);
	}

	if((skb = skb_recv_datagram(sk, flags, noblock, &error) ) == NULL) {
                return error;
	}

	if(size > skb->len) {
		size = skb->len;
	}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
	else if(size <skb->len) {
		msg->msg_flags |= MSG_TRUNC;
	}
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */

	skb_copy_datagram_iovec(skb, 0, msg->msg_iov, size);
        sk->stamp=skb->stamp;

	if(ska) {
		memset(ska, 0, sizeof(* ska));
		ska->sa_family = PF_KEY;
	}

	skb_free_datagram(sk, skb);
	return size;
}

static int
pfkey_shutdown(struct socket *sock, int mode)
{
	pfkey_sock *sk;

	if(sock == NULL) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_shutdown: NULL socket passed in.\n");
		return 0;
	}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
	sk=sock->sk;
#else /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
	sk=sock->data;
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
	
	if(sk == NULL) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_shutdown: No sock attached to socket.\n");
		return 0;
	}

	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_shutdown: .\n");
	mode++;
	
	if(mode&SEND_SHUTDOWN) {
		sk->shutdown|=SEND_SHUTDOWN;
		sk->state_change(sk);
	}

	if(mode&RCV_SHUTDOWN) {
		sk->shutdown|=RCV_SHUTDOWN;
		sk->state_change(sk);
	}
	return 0;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
static int
pfkey_select(struct socket *sock, int sel_type, select_table *wait)
{
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_select: .\n");
	if(sock == NULL) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_select: Null socket passed in.\n");
		return 0;
	}
	if(wait == NULL) {
		KLIPS_PRINT(debug_pfkey,
			    "klips_debug:pfkey_select: Null select_table passed in.\n");
		return 0;
	}
	return datagram_select(sock->data, sel_type, wait);
}
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0) */

static int
pfkey_ioctl(struct socket *sock, unsigned int cmd, unsigned long arg)
{
	KLIPS_PRINT(debug_pfkey,
		    "klips_debug:pfkey_ioctl: not supported.\n");
	return -EINVAL;
}

#ifdef CONFIG_PROC_FS
static int
pfkey_get_info(char *buffer, char **start, off_t offset, int length, int dummy)
{
	off_t pos=0;
	off_t begin=0;
	int len=0;
	pfkey_sock *s=pfkey_sock_list;
	
	len+= sprintf(buffer,"Num       Protocol Flags    Type State\n");
	
	while(s!=NULL) {
		len+=sprintf(buffer+len,"%p: %08lX %08X %02X\n",
			s,
			s->socket->flags,
			s->socket->type,
			s->socket->state);
		
		pos=begin+len;
		if(pos<offset)
		{
			len=0;
			begin=pos;
		}
		if(pos>offset+length)
			break;
		s=s->next;
	}
	*start=buffer+(offset-begin);
	len-=(offset-begin);
	if(len>length)
		len=length;
	return len;
}
#endif /* CONFIG_PROC_FS */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
struct proto_ops pfkey_dgram_ops
#else /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
struct proto_ops pfkey_proto_ops
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
 = {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
	PF_KEY,
	
#else /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
	AF_KEY,
	
	pfkey_create,
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
	pfkey_dup,
	pfkey_release,
	pfkey_bind,
	pfkey_connect,
	pfkey_socketpair,
	pfkey_accept,
	pfkey_getname,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
	datagram_poll,
#else /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
	pfkey_select,
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
	pfkey_ioctl,
	pfkey_listen,
	pfkey_shutdown,
	pfkey_setsockopt,
	pfkey_getsockopt,
	pfkey_fcntl,
	pfkey_sendmsg,
	pfkey_recvmsg
};
   
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
struct net_proto_family pfkey_family_ops = {
	PF_KEY,
	
	pfkey_create
};
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */

#ifdef CONFIG_PROC_FS
struct proc_dir_entry proc_net_pfkey =
{
#ifdef MODULE
	0,
#else /* MODULE */
	PROC_NET_PF_KEY,
#endif /* MODULE */
	6, "pf_key",
	S_IFREG | S_IRUGO, 1, 0, 0,
	0, &proc_net_inode_operations,
	pfkey_get_info
};
#endif /* CONFIG_PROC_FS */

/* void */
int
pfkey_init(void)
{
	int error = 0;
	
        printk(KERN_INFO "FreeS/WAN: initialising PF_KEY domain sockets.\n");
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
        sock_register(&pfkey_family_ops);
#else /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
        sock_register(pfkey_proto_ops.family, &pfkey_proto_ops);
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
#ifdef CONFIG_PROC_FS
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
#ifdef MODULE
        proc_register_dynamic
#else /* MODULE */
	proc_register
#endif /* MODULE */
#else /* LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0) */
	proc_register
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0) */
		(
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
			&
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0) */
			proc_net, 
			&proc_net_pfkey);
#endif /* CONFIG_PROC_FS */
	return error;
}

/* void */
int
pfkey_cleanup(void)
{
	int error = 0;
	
        printk(KERN_INFO "FreeS/WAN: shutting down PF_KEY domain sockets.\n");
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
        sock_unregister(PF_KEY);
#else /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
        sock_unregister(pfkey_proto_ops.family);
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0) */
#ifdef CONFIG_PROC_FS
        if(proc_unregister(
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
		&
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0) */
		proc_net, proc_net_pfkey.low_ino) != 0)
		printk("klips_debug:pfkey_cleanup: cannot unregister
/proc/net/pf_key\n");
	;
#endif /* CONFIG_PROC_FS */

	/* other module unloading cleanup happens here */
	return error;
}

#ifdef MODULE
#if 0
int
init_module(void)
{
	pfkey_init();
	return 0;
}

void
cleanup_module(void)
{
	pfkey_cleanup();
}
#endif /* 0 */
#else /* MODULE */
void
pfkey_proto_init(struct net_proto *pro)
{
	pfkey_init();
}
#endif /* MODULE */

/*
 * $Log: pfkey_v2.c,v $
 * Revision 1.11  1999/09/30 02:55:10  rgb
 * Bogus skb detection.
 * Fix incorrect /proc/net/ipsec-eroute printk message.
 *
 * Revision 1.10  1999/09/21 15:22:13  rgb
 * Temporary fix while I figure out the right way to destroy sockets.
 *
 * Revision 1.9  1999/07/08 19:19:44  rgb
 * Fix pointer format warning.
 * Fix missing member error under 2.0.xx kernels.
 *
 * Revision 1.8  1999/06/13 07:24:04  rgb
 * Add more debugging.
 *
 * Revision 1.7  1999/06/10 05:24:17  rgb
 * Clarified compiler directives.
 * Renamed variables to reduce confusion.
 * Used sklist_*_socket() kernel functions to simplify 2.2.x socket support.
 * Added lots of sanity checking.
 *
 * Revision 1.6  1999/06/03 18:59:50  rgb
 * More updates to 2.2.x socket support.  Almost works, oops at end of call.
 *
 * Revision 1.5  1999/05/25 22:44:05  rgb
 * Start fixing 2.2 sockets.
 *
 * Revision 1.4  1999/04/29 15:21:34  rgb
 * Move log to the end of the file.
 * Eliminate min/max redefinition in #include <net/tcp.h>.
 * Correct path for pfkey #includes
 * Standardise an error return method.
 * Add debugging instrumentation.
 * Move message type checking to pfkey_msg_parse().
 * Add check for errno incorrectly set.
 * Add check for valid PID.
 * Add check for reserved illegally set.
 * Add check for message out of bounds.
 *
 * Revision 1.3  1999/04/15 17:58:07  rgb
 * Add RCSID labels.
 *
 * Revision 1.2  1999/04/15 15:37:26  rgb
 * Forward check changes from POST1_00 branch.
 *
 * Revision 1.1.2.2  1999/04/13 20:37:12  rgb
 * Header Title correction.
 *
 * Revision 1.1.2.1  1999/03/26 20:58:55  rgb
 * Add pfkeyv2 support to KLIPS.
 *
 *
 * RFC 2367
 * PF_KEY_v2 Key Management API
 */
