RCSID $Id: net.ipv4.ip_output.c,v 1.2 1999/08/25 14:35:28 rgb Exp $
--- ./net/ipv4/ip_output.c.preipsec	Thu Mar 25 12:23:34 1999
+++ ./net/ipv4/ip_output.c	Thu Aug 19 03:37:52 1999
@@ -82,6 +82,7 @@
  */
 
 int sysctl_ip_dynaddr = 0;
+int sysctl_ipfrag_debug = 0;
 
 
 int ip_id_count = 0;
@@ -728,14 +729,51 @@
 	struct iphdr *iph;
 	unsigned char *raw;
 	unsigned char *ptr;
-	struct device *dev;
+	struct device *dev = NULL;
 	struct sk_buff *skb2;
 	unsigned int mtu, hlen, left, len; 
 	int offset;
 	int not_last_frag;
-	struct rtable *rt = (struct rtable*)skb->dst;
+	struct rtable *rt;
 
-	dev = rt->u.dst.dev;
+	if(sysctl_ipfrag_debug == 1) {	/* RGB delete section */
+		printk(KERN_DEBUG "ip_fragment:");
+		printk("called, frgdbg=%d skb=%p output=%p\n",
+		       sysctl_ipfrag_debug, skb, output);
+		if(skb) {
+			kfree_skb(skb);
+		}
+		return;
+	}
+	if(!skb) {
+		printk(KERN_INFO "ip_fragment:NULL skb pointer passed in, \n");
+		return;
+	}
+
+	if(!output) {
+		printk(KERN_INFO "ip_fragment:NULL output function pointer passed in, dropped\n");
+		kfree_skb(skb);
+		return;
+	}
+
+	rt = (struct rtable*)skb->dst;
+
+	if(rt) {
+		dev = rt->u.dst.dev;
+	} else {
+		printk(KERN_DEBUG "ip_fragment:no device associated with skb=%p skb->dst=%p rt=%p\n",
+		       skb, skb->dst, rt);
+	}
+
+	if(!skb->dev) {
+		printk(KERN_DEBUG "ip_fragment:no device associated with skb\n");
+	}
+
+	if(!dev && !skb->dev) {
+		printk(KERN_INFO "ip_fragment:no device associated with skb, dropped\n");
+		kfree_skb(skb);
+		return;
+	}
 
 	/*
 	 *	Point into the IP datagram header.
@@ -748,9 +786,26 @@
 	 *	Setup starting values.
 	 */
 
+	if(sysctl_ipfrag_debug == 2) {	/* RGB delete section */
+		printk(KERN_DEBUG "ip_fragment:");
+		printk("called, frgdbg=%d skb=%p skb->dst=%p skb->dev=%p output=%p raw=%p\n",
+		       sysctl_ipfrag_debug, skb, skb->dst, skb->dev, output, raw);
+		kfree_skb(skb);
+		return;
+	}
+	if(!raw) {
+		goto fail;
+	}
+
 	hlen = iph->ihl * 4;
 	left = ntohs(iph->tot_len) - hlen;	/* Space per frame */
-	mtu = rt->u.dst.pmtu - hlen;	/* Size of data space */
+	if((!(skb->dev) || (dev == skb->dev)) && rt) { /* RGB */
+		mtu = rt->u.dst.pmtu - hlen;	/* Size of data space */
+	} else {
+		mtu = skb->dev->mtu - hlen;
+		dev = skb->dev;
+	}
+
 	ptr = raw + hlen;			/* Where to start from */
 
 	/*
@@ -773,10 +828,80 @@
 	offset = (ntohs(iph->frag_off) & IP_OFFSET) << 3;
 	not_last_frag = iph->frag_off & htons(IP_MF);
 
+	if(sysctl_ipfrag_debug) {	/* RGB delete section */
+		printk(KERN_DEBUG "ip_fragment:");
+		if(skb->dst) {
+			if(skb->dst->dev) {
+				if(skb->dst->dev->name) {
+					printk("skb->dst->dev=%s", skb->dst->dev->name);
+				}
+				printk(" skb->dst->dev->hhlen:%d", skb->dst->dev->hard_header_len);
+				printk(" skb->dst->dev->mtu=%d", skb->dst->dev->mtu);
+			} else {
+				printk(" skb->dst->dev=NULL");
+			}
+		} else {
+			printk(" skb->dst=NULL");
+		}
+		if(dev) {
+			if(dev->name) {
+				printk(" dev=%s", dev->name);
+			}
+			printk(" dev->hhlen:%d", dev->hard_header_len);
+			printk(" dev->mtu=%d", dev->mtu);
+		} else {
+			printk(" dev=NULL");
+		}
+		printk(" mtu=%d", mtu);
+		printk(" left=%d", left);
+		printk("\n");
+	if(sysctl_ipfrag_debug == 3) {	/* RGB delete section */
+		printk(KERN_DEBUG "ip_fragment:");
+		printk("called, frgdbg=%d output=%p\n",
+		       sysctl_ipfrag_debug, output);
+		kfree_skb(skb);
+		return;
+	}
+		printk(KERN_DEBUG "ip_fragment:");
+		printk(" id:%d", iph->id);
+		printk(" tlen:%d", ntohs(iph->tot_len));
+		printk(" ihl:%d", hlen);
+		printk(" offset:%d", offset);
+		if(ntohs(iph->frag_off) & IP_MF)
+			printk(" MF");
+		if(ntohs(iph->frag_off) & IP_DF)
+			printk(" DF");
+		printk(" %ld.%ld.%ld.%ld",
+		       ntohl(iph->saddr) >> 24 & 0xff,
+		       ntohl(iph->saddr) >> 16 & 0xff,
+		       ntohl(iph->saddr) >> 8 & 0xff,
+		       ntohl(iph->saddr) >> 0 & 0xff);
+		printk("->%ld.%ld.%ld.%ld",
+		       ntohl(iph->daddr) >> 24 & 0xff,
+		       ntohl(iph->daddr) >> 16 & 0xff,
+		       ntohl(iph->daddr) >> 8 & 0xff,
+		       ntohl(iph->daddr) >> 0 & 0xff);
+		printk(" proto:%d", iph->protocol);
+#if 0
+		printk(" ttl:%d", iph->ttl);
+		printk(" ver:%d", iph->version);
+		printk(" tos:%d", iph->tos);
+		printk(" chk:%d", iph->check);
+#endif /* 0 */
+		printk("\n");
+	}
+
 	/*
 	 *	Keep copying data until we run out.
 	 */
 
+	if(sysctl_ipfrag_debug == 4) {	/* RGB delete section */
+		printk(KERN_DEBUG "ip_fragment:");
+		printk("called, frgdbg=%d output=%p\n",
+		       sysctl_ipfrag_debug, output);
+		kfree_skb(skb);
+		return;
+	}
 	while(left > 0)	{
 		len = left;
 		/* IF: it doesn't fit, use 'mtu' - the data space left */
@@ -800,9 +925,25 @@
 		 *	Set up data on packet
 		 */
 
+		if(sysctl_ipfrag_debug == 5) {	/* RGB delete section */
+			printk(KERN_DEBUG "ip_fragment:");
+			printk("called, frgdbg=%d output=%p\n",
+			       sysctl_ipfrag_debug, output);
+			kfree_skb(skb);
+			kfree_skb(skb2);
+			return;
+		}
 		skb2->pkt_type = skb->pkt_type;
 		skb2->priority = skb->priority;
 		skb_reserve(skb2, (dev->hard_header_len+15)&~15);
+		if(sysctl_ipfrag_debug == 6) {	/* RGB delete section */
+			printk(KERN_DEBUG "ip_fragment:");
+			printk("called, frgdbg=%d output=%p\n",
+			       sysctl_ipfrag_debug, output);
+			kfree_skb(skb);
+			kfree_skb(skb2);
+			return;
+		}
 		skb_put(skb2, len + hlen);
 		skb2->nh.raw = skb2->data;
 		skb2->h.raw = skb2->data + hlen;
@@ -812,6 +953,14 @@
 		 *	it might possess
 		 */
 
+		if(sysctl_ipfrag_debug == 7) {	/* RGB delete section */
+			printk(KERN_DEBUG "ip_fragment:");
+			printk("called, frgdbg=%d output=%p\n",
+			       sysctl_ipfrag_debug, output);
+			kfree_skb(skb);
+			kfree_skb(skb2);
+			return;
+		}
 		if (skb->sk)
 			skb_set_owner_w(skb2, skb->sk);
 		skb2->dst = dst_clone(skb->dst);
@@ -840,6 +989,14 @@
 		 * on the initial skb, so that all the following fragments
 		 * will inherit fixed options.
 		 */
+		if(sysctl_ipfrag_debug == 8) {	/* RGB delete section */
+			printk(KERN_DEBUG "ip_fragment:");
+			printk("called, frgdbg=%d output=%p\n",
+			       sysctl_ipfrag_debug, output);
+			kfree_skb(skb);
+			kfree_skb(skb2);
+			return;
+		}
 		if (offset == 0)
 			ip_options_fragment(skb);
 
@@ -861,7 +1018,30 @@
 		iph->tot_len = htons(len + hlen);
 
 		ip_send_check(iph);
-
+		
+		if(sysctl_ipfrag_debug) {	/* RGB delete section */
+			printk(KERN_DEBUG "ip_fragment:  ");
+			printk(" tlen:%d", ntohs(iph->tot_len));
+			printk(" ihl:%d", hlen);
+			printk(" left=%d", left);
+			printk(" len=%d", len);
+			printk(" offset:%d", (ntohs(iph->frag_off) & IP_OFFSET) << 3);
+			if(ntohs(iph->frag_off) & IP_MF)
+				printk(" MF");
+			if(ntohs(iph->frag_off) & IP_DF)
+				printk(" DF");
+/*			printk(" chk:%d", iph->check); */
+			printk("\n");
+		}
+			
+		if(sysctl_ipfrag_debug == 9) {	/* RGB delete section */
+			printk(KERN_DEBUG "ip_fragment:");
+			printk("called, frgdbg=%d output=%p\n",
+			       sysctl_ipfrag_debug, output);
+			kfree_skb(skb);
+			kfree_skb(skb2);
+			return;
+		}
 		output(skb2);
 	}
 	kfree_skb(skb);
